#!/usr/bin/python3
import unittest

import initdutils as iu

class TestInitdUtils(unittest.TestCase):

	@unittest.skip('Test not implemented.')
	def test_scan_initfile():
		raise NotImplementedError()
	@unittest.skip('Test not implemented.')
	def test_save_facilities():
		raise NotImplementedError()
	@unittest.skip('Test not implemented.')
	def test_load_facilities():
		raise NotImplementedError()
	@unittest.skip('Test not implemented.')
	def test_load_depends():
		raise NotImplementedError()
	@unittest.skip('Test not implemented.')
	def test_save_depends():
		raise NotImplementedError()
	@unittest.skip('Test not implemented.')
	def test_load_lsbinstall_info():
		raise NotImplementedError()
	@unittest.skip('Test not implemented.')
	def test_save_lsbinstall_info():
		raise NotImplementedError()

if __name__ == '__main__':
	unittest.main()
